import numpy as np

class base_sampler(object):
    def __init__(self, data_source):
        pass

    def __iter__(self):
        raise NotImplementedError

    def __len__(self):
        raise NotImplementedError

class sequential_sampler(base_sampler):
    """return inidices in orginal order"""
    def __init__(self, data_source):
        self.data_source = data_source
    
    def __iter__(self):
        return iter(range(len(self.data_source)))

    def __len__(self):
        return len(self.data_source)

class random_sampler(base_sampler):
    """shuffles indicies for sampling"""
    def __init__(self, data_source):
        self.data_source = data_source
    def __iter__(self):
        return iter(np.random.permutation(len(self.data_source)))
    def __len__(self):
        return len(self.data_source)
    
class minibatch_sampler(base_sampler):
    def __init__(self, sampler, batch_size, drop_last):
        self.sampler = sampler
        self.batch_size = batch_size
        self.drop_last = drop_last

    def __iter__(self):
        batch = []
        for idx in self.sampler:
            batch.append(idx)
            if len(batch) == self.batch_size:
                yield batch
                batch = []
        if len(batch) > 0 and not self.drop_last:
            yield batch
    
    def __len__(self):
        if self.drop_last:
            return len(self.sampler) // self.batch_size
        else:
            return (len(self.sampler) + self.batch_size - 1) // self.batch_size
        




# Copyright 2015 Jeff Vierstra

import numpy as np

class genomic_interval(object):
	"""Class that implements BED-style object
	"""

	def __init__(self, chrom, start, end, name='.', score=None, strand=None, extra=None):
		self.chrom = str(chrom)
		self.start = int(start)
		self.end = int(end)
		self.name = str(name)
		self.score = score
		self.strand = strand
		self.extra = extra

	def __len__(self):
		"""Length of element
		"""
		return self.end - self.start
	
	def __str__(self):
		"""Returns a string-formated version of the element 
		for printing
		"""
		return '\t'.join( [ str(x) for x in [self.chrom, self.start, self.end] ] )

	def to_ucsc(self):
		"""Prints out a UCSC version of interval. Note
		that the start coordinate is shifted by 1 to
		reflect 1-based"""
		return f"{self.chrom}:{self.start}-{self.end}"

	def widen(self, x, inplace=False):
		"""Expands the coordinates
		"""
		if inplace:
			self.start -= x
			self.end += x
			return self
		return genomic_interval(self.chrom, self.start-x, self.end+x, self.name, self.strand)

	def shift(self, x, inplace=False):
		"""Shifts the coordinates"""
		if inplace:
			self.start += x
			self.end += x
			return self
		return genomic_interval(self.chrom, self.start+x, self.end+x, self.name, self.strand)




import matplotlib as mpl
import matplotlib.colors as mcolors
import matplotlib.cm as mcm

from cycler import cycler
from itertools import cycle

import numpy as np

from .cm import COLOR_MAPS

__all__ = ["get_color_cycle", "set_palette", "color_palette"]

QUAL_PALETTES = [
    "zissou", "darjeeling", "lawhoops",
]
QUAL_PALETTES_SIZES = {name: len(COLOR_MAPS[name]) for name in QUAL_PALETTES}

def get_color_cycle():
    """Returns the current color cycle from matplotlib prop cycler"""
    cyl = mpl.rcParams['axes.prop_cycler']
    return cyl.by_key()['color'] if 'color' in cyl.keys else ['k']

class _ColorPalette(list):
    """Enables color palette used with `with` command"""
    def __enter__(self):
        self.orig_palette = color_palette()
        set_palette(self)
        return self

    def __exit__(self):
        set_palette(self.orig_palette)

def mpl_cmap_palette(cmap, n_colors, as_cmap=False):
    """Returns N colors (evenly distributed) from a colormap"""

    if isinstance(cmap, str):
        cmap = mcm.get_cmap(cmap)

    bins = np.linspace(0, 1, int(n_colors) + 2)[1:-1]
    colors = list(map(tuple, cmap(bins)[:,:3]))

    if as_cmap:
        return cmap
    else:
        return _ColorPalette(colors)

def discrete_cmap(cmap, n_colors):
    """Create an N-bin discrete colormap from the specified input map"""

    # Note that if base_cmap is a string or None, you can simply do
    #    return plt.cm.get_cmap(base_cmap, N)
    # The following works for string, None, or a colormap instance:

    if isinstance(cmap, str):
        cmap = mcm.get_cmap(cmap)
        
    color_list = mpl_cmap_palette(cmap, n_colors, as_cmap=False)
    return mcolors.ListedColormap(color_list)

def color_palette(palette=None, n_colors=None, as_cmap=False):

    if palette is None:
        palette = get_color_cycle()
        if n_colors is None:
            n_colors = len(palette)

    elif not isinstance(palette, str):
        palette = palette
        if n_colors is None:
            n_colors = len(palette)
    
    else:
        if n_colors is None:
            n_colors = QUAL_PALETTES_SIZES.get(palette, 6)

        if palette in QUAL_PALETTES:
            palette = COLOR_MAPS[palette]
        else:
            try:
                palette = mpl_cmap_palette(palette, n_colors, as_cmap=as_cmap)
            except:
                raise ValueError(f'{palette} is not a valid palette name.')

    if not as_cmap:

        palette_cycle = cycle(palette)
        palette = [next(palette_cycle) for _ in range(n_colors)]   

        try:
            palette = map(mcolors.to_rbg, palette)
            palette = _ColorPalette(palette)
        except ValueError:
            raise ValueError(f'Could not generate a palette for {palette}. Check color formats')

    return palette

def set_palette(palette, n_colors):
    """Sets the palette to the default matplotlib prop cycler"""

    colors = color_palette(palette, n_colors)
    cyl = cycler('color', colors)
    mpl.rcParams['axes.prop_cycle'] = cyl
    mpl.rcParams['patch.facecolor'] = colors[0]



"""Utilities for sequence manipulation"""

import numpy as np

IN_MAP = np.asarray([[0, 0, 0, 0],
                    [1, 0, 0, 0],
                    [0, 1, 0, 0],
                    [0, 0, 1, 0],
                    [0, 0, 0, 1]], dtype='int8')

def get_base_encoding():
    base_encoding = np.zeros((ord('t')+1, 1), dtype=np.int)
    base_encoding[ord('A')] = 1
    base_encoding[ord('C')] = 2
    base_encoding[ord('G')] = 3
    base_encoding[ord('T')] = 4
    base_encoding[ord('a')] = 1
    base_encoding[ord('c')] = 2
    base_encoding[ord('g')] = 3
    base_encoding[ord('t')] = 4
    return base_encoding

def one_hot_encode(seqs, revcomp=False):
    base_encoding = get_base_encoding()
    if isinstance(seqs, str) or isinstance(seqs, bytes):
        seqs = [seqs]
    int_seqs = np.concatenate([base_encoding[list(bytearray(str(s), encoding='utf8'))] for s in seqs], axis=1)
    out = IN_MAP[int_seqs]
    if revcomp:
        out = np.flip(np.flip(out, axis=0), axis=1)
    return out


"""Module to facilitate the loading a procesing of data. Largely
inspired/copied from Kipoi (http://github.com/kipoi) and pyTorch
data loaders"""

__all__ = ["extractors", "dataset", "loaders", "samplers", "utils"]


import matplotlib.pyplot as plt
import matplotlib.lines as mlines

import numpy as np

import pysam

from genome_tools import genomic_interval

from .utils import pack_rows
from .colors.cm import get_vocab_color

ANNOTATION_KWDS = {
    'exon': {},
    'transcript': {},
    'utr': {}
}

def parse_gff_attribs(s):
    attr = {}
    for field in s.split(';'):
        (key, val) = field.split('=')
        attr[key] = val
    return attr

def load_gff_data(interval, gff_file):
    
    genes = {}
    transcripts = {}
    exons = {}
    cdss = {}
    utrs = {}

    fh = pysam.TabixFile(gff_file)
    
    for row in fh.fetch(interval.chrom, interval.start, interval.end, parser=pysam.asGTF()):
        
        feature_interval = genomic_interval(row.contig, row.start, row.end, strand = row.strand)
        attribs = parse_gff_attribs(row.attributes)

        if row.feature == 'gene':
            genes[attribs['gene_id']] = attribs['gene_name']
        elif row.feature == 'transcript':
           transcripts[attribs['gene_id']] = transcripts.get(attribs['gene_id'], []) + [(attribs['transcript_id'], feature_interval)]
        elif row.feature == 'exon':
            exons[attribs['transcript_id']] = exons.get(attribs['transcript_id'], []) + [(attribs['exon_id'], feature_interval)]
        elif row.feature == 'CDS':
            cdss[attribs['exon_id']] = cdss.get(attribs['exon_id'], []) + [(attribs['ID'], feature_interval)]
        elif row.feature == 'UTR':
            utrs[attribs['exon_id']] = utrs.get(attribs['exon_id'], []) + [(attribs['ID'], feature_interval)]

    fh.close()

    return genes, transcripts, exons, cdss, utrs


# Todo add color as styling kwargs
def gene_annotation_plot(interval, annotation_file, ax=None):

    if ax is None:
        ax = plt.gca()
    figure = ax.get_figure()

    genes, transcripts, exons, cdss, utrs = load_gff_data(interval, annotation_file)

    t = []
    for gene_id, gene_transcripts in transcripts.items():
        for transcript_id, transcript in gene_transcripts:
            t.append(transcript)

    row_indices = pack_rows(t)

    label_xoffset = len(interval) * 0.04

    label_arrow_height = 0.3
    label_arrow_width = len(interval) * 0.025
    label_arrow_xoffset = 0
    label_arrow_yoffset = 0.2

    for gene_id, gene_symbol in genes.items():
        
        if gene_id not in transcripts:
            continue

        for transcript_id, transcript in transcripts.get(gene_id, []):
            
            i = row_indices[transcript]

            over_left = over_right = start_visible = False

            if transcript.start < interval.start:
                over_left = True
                start = interval.start
            else:
                start = transcript.start

            if transcript.end > interval.end:
                over_right = True
                end = interval.end
            else:
                end = transcript.end

            # Draw trasnscript line
            ax.plot([start, end], [i, i], color='k', lw=2)

            for exon_id, exon in exons.get(transcript_id, []):
                
                for utr_id, utr in utrs.get(exon_id, []):
                    # make this a rectangle so we can put a border
                    l = mlines.Line2D([utr.start, utr.end], [i, i], color='k', lw=4)
                    ax.add_line(l)

                for cds_id, cds in cdss.get(exon_id, []):
                    l = mlines.Line2D([cds.start, cds.end], [i, i], color='blue', lw=5)
                    ax.add_line(l)

            # Draw labels/arrrow/etc...
            if transcript.strand == '+':

                label_arrow_xcoords = np.array([start, start, start + label_arrow_width])
                label_arrow_ycoords = np.array([i, i + label_arrow_height, i + label_arrow_height]) + label_arrow_yoffset
                ma = '>'

                if not over_left:
                    start_visible = True
                    label_xcoord = start + label_xoffset
                    label_ycoord = i + label_arrow_height + label_arrow_yoffset
                    ha = 'left'
                else:
                    label_xcoord = start - label_xoffset
                    label_ycoord = i
                    ha = 'right'

            else: 

                label_arrow_xcoords = np.array([end, end, end - label_arrow_width])
                label_arrow_ycoords = np.array([i, i + label_arrow_height, i + label_arrow_height]) + label_arrow_yoffset
                ma = '<'

                if not over_right:
                    start_visible = True
                    label_xcoord = end - label_xoffset
                    label_ycoord = i + label_arrow_height + label_arrow_yoffset
                    ha = 'right'
                else:
                    label_xcoord = end + label_xoffset
                    label_ycoord = i
                    ha = 'left'

            if start_visible:
                ax.plot(label_arrow_xcoords, label_arrow_ycoords, color = 'black')
                ax.plot(label_arrow_xcoords[-1], label_arrow_ycoords[-1], marker = ma, color = 'black', markeredgecolor='none')

            ax.text(label_xcoord, label_ycoord, gene_symbol, style = 'italic', verticalalignment = 'center', horizontalalignment = ha)

    ax.set_xlim(interval.start, interval.end)
    ax.set_ylim(bottom=-0.5, top=max(row_indices.values())+1)
    ax.yaxis.set_visible(False)


import numpy as np
import pandas as pd

from collections.abc import Iterable, Sequence, Mapping

from matplotlib.projections import register_projection

import matplotlib.axes as maxes
import matplotlib.axis as maxis
import matplotlib.collections as mcollections
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import matplotlib.text as mtext
import matplotlib.transforms as mtransforms
import matplotlib.ticker as mticker

from mpl_toolkits.axes_grid1.inset_locator import inset_axes, InsetPosition
from mpl_toolkits.axes_grid1.anchored_artists import AnchoredSizeBar

# registers custom colormaps
# from .colors import cm
from .utils import (rescale_data, 
                    format_axes_to_interval,
                    clear_spines, pack_rows)

from .gene_annotation import gene_annotation_plot

class AxisAnnotation(mtext.Annotation):
    def __init__(self, text, xy, xytext, xycoords, **kwargs):
        super().__init__(text, xy, xytext=xytext, xycoords=xycoords, **kwargs)

    def get_text_window_extent(self, renderer):
        if self.get_bbox_patch():
            return self.get_bbox_patch().get_window_extent(renderer)
        else:
            return mtext.Text.get_window_extent(self)

    def update_positions(self, renderer):
        """
        """
        x1, y1 = self._get_position_xy(renderer)  # Annotated position.
        # generate transformation,
        self.set_transform(self._get_xy_transform(renderer, self.anncoords))

        if self.arrowprops is None:
            return

        bbox = mtext.Text.get_window_extent(self, renderer)

        d = self.arrowprops.copy()

        ms = d.pop("mutation_scale", self.get_size())
        self.arrow_patch.set_mutation_scale(ms)

        shrink = d.pop('shrink', 0.0)

        if "arrowstyle" not in d:
            self.arrow_patch.set_arrowstyle('-')


        # Pick the corner of the text bbox closest to annotated point.
        xpos = [(bbox.x0, 0), ((bbox.x0 + bbox.x1) / 2, 0.5), (bbox.x1, 1)]
        ypos = [(bbox.y0, 0), ((bbox.y0 + bbox.y1) / 2, 0.5), (bbox.y1, 1)]

        xypos = []

        for i in xpos:
            for j in ypos:
                xypos.append((i, j))

        xmin, ymin = min(xypos, key=lambda v: abs(np.hypot(v[1][0] - y1, v[0][0] - x1)))

        relposx, relposy = xmin[1], ymin[1]
        x, y = xmin[0], ymin[0]

        # x, relposx = min(xpos, key=lambda v: abs(v[0] - x1))
        # y, relposy = min(ypos, key=lambda v: abs(v[0] - y1))

        self._arrow_relpos = (relposx, relposy)
        r = np.hypot(y - y1, x - x1)
        shrink_pts = shrink * r / renderer.points_to_pixels(1)
        self.arrow_patch.shrinkA = self.arrow_patch.shrinkB = shrink_pts

        # adjust the starting point of the arrow relative to the textbox.
        # TODO : Rotation needs to be accounted.
        relposx, relposy = self._arrow_relpos
        x0 = bbox.x0 + bbox.width * relposx
        y0 = bbox.y0 + bbox.height * relposy

        # The arrow will be drawn from (x0, y0) to (x1, y1). It will be first
        # clipped by patchA and patchB.  Then it will be shrunk by shrinkA and
        # shrinkB (in points).  If patch A is not set, self.bbox_patch is used.
        self.arrow_patch.set_positions((x0, y0), (x1, y1))

        if "patchA" in d:
            self.arrow_patch.set_patchA(d.pop("patchA"))
        else:
            if self._bbox_patch:
                self.arrow_patch.set_patchA(self._bbox_patch)
            else:
                if self.get_text() == "":
                    self.arrow_patch.set_patchA(None)
                    return
                pad = renderer.points_to_pixels(4)
                r = mpatches.Rectangle(xy=(bbox.x0 - pad / 2, bbox.y0 - pad / 2),
                              width=bbox.width + pad, height=bbox.height + pad,
                              transform=mtransforms.IdentityTransform(), clip_on=False)
                self.arrow_patch.set_patchA(r)

# ------------------------

class AxisAnnotatedRepositionable(object):
    def __init__(self, pad=0):
        self.pad = pad
        self.annotations = []

    def add_annotation(self, text, pos, offset, loc, **kwargs):
        raise NotImplementedError

    def get_annotations(self):
        return self.annotations

    def get_children(self):
        return super().get_children() + [*self.annotations]

    def get_tightbbox(self, renderer, *, for_layout_only=False):
        
        bboxes = [a.get_window_extent(renderer) for a in self.annotations]
        parent_bboxes = super().get_tightbbox(renderer, for_layout_only=for_layout_only)

        if parent_bboxes:
            bboxes.append(parent_bboxes)

        if bboxes:
            return mtransforms.Bbox.union(bboxes)
        else:
            return None

        def _get_annotation_bboxes(self, renderer):
            return [a.get_window_extent(renderer) for a in self.annotations if a.get_visible()]


    def _reposition_axis_annotations(self, renderer):
        
        order = [x for x,y in sorted(enumerate(self.annotations), key=lambda z: z[1].xy[0])]
        
        mid = len(order)//2
        shift_order = order[:mid][::-1] + order[mid:]

        seen_bboxes = []

        for i, idx in enumerate(shift_order):
            current_artist = self.annotations[idx]
            current_bbox = current_artist.get_text_window_extent(renderer)
        
            if len(seen_bboxes) == 0:
                seen_bboxes.append(current_bbox)
                continue
            
            if self.axis_name == 'x':
                slide = (1,0) if i >= mid else (0, 0)
            elif self.axis_name == 'y':
                slide = (1,1) if i >= mid else (0, 1)
            else:
                raise ValueError(f'Axis {self.axis_name} not supported!')

            slide_dir  = int((slide[0] - 0.5)*2)
            furthest = -slide_dir * float('inf')
            shift = 0
            overlaps = False

            for bbox in seen_bboxes:

                if current_bbox.overlaps(bbox):
                    overlaps = True

                bbox_pt = bbox.get_points()[slide]
                current_bbox_pt = current_bbox.get_points()[1-slide[0], slide[1]]

                if slide_dir * bbox_pt > slide_dir * furthest:
                    furthest = bbox_pt
                    shift = slide_dir * (furthest - current_bbox_pt)
            
            if shift and overlaps:
                xy = np.array(current_artist.get_position())
                xy[slide[1]] += (shift + self.pad) * slide_dir *  (72. / self.figure.get_dpi())
                current_artist.set_position(xy)
                current_artist.update_bbox_position_size(renderer)

            seen_bboxes.append(current_artist.get_text_window_extent(renderer))
    
    def draw(self, renderer, *args, **kwargs):
        super().draw(renderer, *args, **kwargs)

        for a in self.annotations:
            a.update_positions(renderer)
            a.update_bbox_position_size(renderer)
                
        self._reposition_axis_annotations(renderer)

        renderer.open_group(__name__, gid=self.get_gid())

        for a in self.annotations:
            a.draw(renderer)

        renderer.close_group(__name__)

# ------------------------

class XAxisAnnotated(maxis.XAxis, AxisAnnotatedRepositionable):
    def __init__(self, *args, **kwargs):
        maxis.XAxis.__init__(self, *args, **kwargs)
        AxisAnnotatedRepositionable.__init__(self, pad=5)
  
    def add_annotation(self, text, pos, offset=15, loc='bottom', **kwargs):
        """
        x: data coordinates

        """
        if loc == 'top':
            y = 1
        elif loc == 'bottom':
            y = 0
            offset = -offset
        else:
            raise ValueError('`loc` argument must be either top or bottom')

        va = kwargs.pop('va', 'bottom' if loc == 'top' else 'top')
        ha = kwargs.pop('ha', 'center')

        arrowprops = kwargs.pop('arrowprops', dict(arrowstyle='-'))

        t = AxisAnnotation(
            text=text,
            xy=(pos, y), 
            xytext=(0, offset), 
            xycoords=self.axes.get_xaxis_transform('tick1'), 
            textcoords='offset points', 
            ha=ha, va=va, 
            arrowprops=arrowprops,
            **kwargs)
        
        t.set_figure(self.axes.figure)
        self.annotations.append(t)

# ------------------------

class YAxisAnnotated(maxis.YAxis, AxisAnnotatedRepositionable):
    def __init__(self, *args, **kwargs):
        maxis.YAxis.__init__(self, *args, **kwargs)
        AxisAnnotatedRepositionable.__init__(self, pad=5)
  
    def add_annotation(self, text, pos, offset=15, loc='right', **kwargs):
        """
        x: data coordinates

        """
        if loc == 'left':
            x = 0
            offset = -offset
        elif loc == 'right':
            x = 1
        else:
            raise ValueError('`loc` argument must be either left or right')

        va = kwargs.pop('va', 'center')
        ha = kwargs.pop('ha', 'right' if loc == 'left' else 'left')

        arrowprops = kwargs.pop('arrowprops', dict(arrowstyle='-'))

        t = AxisAnnotation(
            text=text, 
            xy=(x, pos), 
            xytext=(offset, 0), 
            xycoords=self.axes.get_yaxis_transform('tick1'), 
            textcoords='offset points', 
            horizontalalignment=ha, verticalalignment=va, 
            arrowprops=arrowprops,
            **kwargs)
        
        t.set_figure(self.axes.figure)
        self.annotations.append(t)

# ------------------------

class GenomeAxes(plt.Axes):
    name = 'genome'

    def _init_axis(self):
        self.xaxis = XAxisAnnotated(self)
        self.spines.bottom.register_axis(self.xaxis)
        self.spines.top.register_axis(self.xaxis)
        self.yaxis = YAxisAnnotated(self)
        self.spines.left.register_axis(self.yaxis)
        self.spines.right.register_axis(self.yaxis)
        self._update_transScale()

register_projection(GenomeAxes)
GenomeAxesSubplot = maxes.subplot_class_factory(GenomeAxes)
GenomeAxesSubplot.__module__ = GenomeAxes.__module__

# ------------------------

class PlotData:

    semantics = "x", "y", "hue", "style"
    flat_structure = {"x": "@index", "y": "@values"}
    wide_structure = {
        "x": "@index", "y": "@values", "hue": "@columns", "style": "@columns",
    }

    def __init__(self, data=None, variables={}):
        self.assign_variables(data, variables)

    @classmethod
    def get_semantics(cls, kwargs, semantics=None):
        if semantics is None:
            semantics = cls.semantics
        variables = {}
        for key, val in kwargs.items():
            if key in semantics and val is not None:
                variables[key] = val
        return variables
         
    def assign_variables(self, data=None, variables={}):
        x = variables.get('x', None)
        y = variables.get('y', None)

    def _assign_variables_wideform(self, data=None, **kwargs):
        assigned = [k for k, v in kwargs.items() if v is not None]
        if any(assigned):
            raise ValueError('')

        empty = data is None or not len(data)

        if isinstance(data, dict):
            values = data.values()
        else:
            values = np.atleast_1d(np.asarray(data, dtype=object))
        flat = not any(
            isinstance(v, Iterable) and not isinstance(v, (str, bytes))
            for v in values)

        if empty:
            plot_data = pd.DataFrame()
            variables = {}

        elif flat:
            flat_data = pd.Series(data).copy()
            names = {
                '@values': flat_data.name,
                '@index': flat_data.index.name
            }

            plot_data = {}
            variables = {}

            for var in ['x', 'y']:
                if var in self.flat_structure:
                    attr = self.flat_structure[var]
                    plot_data[var] = getattr(flat_data, attr[1:])
                    variables = names[self.flat_structure[var]]

            plot_data = pd.DataFrame(plot_data)
        else:
            if isinstance(data, Sequence):
                data_dict = {}
                for i, var in enumerate(data):
                    key = getattr(var, 'name', i)
                    data_dict[key] = pd.Series(var)
                data = data_dict

            if isinstance(data, Mapping):
                data = {key: pd.Series(val) for key, val in data.items()}

            wide_data = pd.DataFrame(data, copy=True)

            melt_kws = {'var_name': '@columns', 'value_name': '@values'}
            use_index = '@index' in self.wide_structure.values()
            if use_index:
                melt_kws['id_vars'] = '@index'
                try:
                    orig_categories = wide_data.columns.categories
                    orig_ordered = wide_data.columns.ordered
                    wide_data.columns = wide_data.columns.add_categories('@index')
                except AttributeError:
                    category_columns = False
                else:
                    category_columns = True

                wide_data['@index'] = wide_data.index.to_series()
            
            plot_data = wide_data.melt(**melt_kws)

            print(plot_data)

            if use_index and category_columns:
                plot_data['@columns'] = pd.Categorical(plot_data['@columns'],
                                                        orig_categories,
                                                        orig_ordered)
            
            for var, attr in self.wide_structure.items():
                plot_data[var] = plot_data[attr]

            variables = {}
            for var, attr in self.wide_structure.items():
                obj = getattr(wide_data, attr[1:])
                variables[var] = getattr(obj, 'name', None)

            plot_data = plot_data[list(variables)]

        return plot_data, variables
        

    def _assign_variables_longform(self, data=None, **kwargs):
      
        plot_data = {}
        variables = {}

        if data is None:
            data = {}

        try:
            index = data.index.to_frame()
        except AttributeError:
            index = {}

        for key, val in kwargs.items():

            try:
                val_as_data_key = (
                    val in data
                    or (isinstance(val, (str, bytes)) and val in index)
                )
            except (KeyError, TypeError):
                val_as_data_key = False

            if val_as_data_key:

                if val in data:
                    plot_data[key] = data[val]
                elif val in index:
                    plot_data[key] = index[val]
                variables[key] = val

            elif isinstance(val, (str, bytes)):
                raise ValueError(f'Could not interpret value `{val}` for parameter `{key}`')

            else:

                if isinstance(data, pd.DataFrame) and not isinstance(val, pd.Series):
                    if np.ndim(val) and len(data) != len(val):
                        val_cls = val.__class__.__name__
                        raise ValueError( 
                            f'Length of {val_cls} vectors must match length of `data`'
                            f' when both are used, but `data` has length {len(data)}'
                            f' and the vector passed to `{key}` has length {len(val)}.')

                plot_data[key] = val

                # Try to infer the name of the variable
                variables[key] = getattr(val, "name", None)

        # Construct a tidy plot DataFrame. This will convert a number of
        # types automatically, aligning on index in case of pandas objects
        plot_data = pd.DataFrame(plot_data)

        # Reduce the variables dictionary to fields with valid data
        variables = {
            var: name
            for var, name in variables.items()
            if plot_data[var].notnull().any()
        }

        return plot_data, variables
# ------------------------


def add_scale_bar(ax):
    scale=100
    bar = AnchoredSizeBar(ax.transData,
            scale, 
            label=f"{scale}nt",
            loc='lower right',
            frameon=False,
            pad=0.4,
            borderpad=1,
            label_top=True,
            fontproperties=dict(size='x-small'))

    bar.patch.set_facecolor('w')
    bar.patch.set_edgecolor('none')
    bar.patch.set_alpha(0.7)
    bar.patch.set_visible(True)
    ax.add_artist(bar)

# ------------------------

def signal_plot(interval, 
                data, 
                ax=None, 
                **kwargs):
    
    if not ax:
        ax = plt.gca()
    fig = ax.get_figure()

    x, y = rescale_data(interval, data, ax, downsample=kwargs.pop('downsample', 0))

    ax.fill_between(x, 0, y, step='mid', **kwargs)

    ax.set_ylim(bottom=0)
    
    format_axes_to_interval(ax, interval)

    return ax
# ------------------------

def grouped_heatmap_plot(interval, 
                         data, 
                         group_level=None, 
                         sort_col=None, 
                         hspace=0.01,
                         cmap='wolfgang_extra',
                         ax=None, 
                         cb_loc='bottom', 
                         cb_label='',
                         **kwargs):

    if not ax:
        ax = plt.gca()
    fig = ax.get_figure()

    if sort_col is not None:
        df = data.sort_values(by=sort_col)
    else:
        df = data.copy()

    group_sizes = data.groupby(level=group_level).size()

    height_ratios = group_sizes / np.sum(group_sizes)
    total = np.sum(height_ratios) + (hspace * (len(group_sizes)-1)) + 0.15

    height_ratios /= total

    cb_height = 0.15/total
    hspace /= total

    ax.set_xlim(interval.start, interval.end)
    ax.set_ylim(0, 1)

    if cb_loc == 'top':
        axes_pos = 0
        cb_pos = 1.0-cb_height
    else:
        axes_pos = cb_height
        cb_pos = 0

    ticks = []
    ticklabels = []

    for group, group_df in data.groupby(level=group_level):
        
        inset_ax = fig.add_axes([0, 0, 1, 1])
        
        ip = InsetPosition(ax, [0, axes_pos, 1, height_ratios[group]]) #posx, posy, width, height
        inset_ax.set_axes_locator(ip)
        
        hm = inset_ax.pcolormesh(group_df, cmap=cmap)

        inset_ax.xaxis.set_visible(False)
        inset_ax.yaxis.set_visible(False)    

        mid = axes_pos + (height_ratios[group]/2)
        ticks.append(mid)
        ticklabels.append(group)

        axes_pos += height_ratios[group]+hspace

    # reserve a portion of ax for colorbar
    cb_bbox_ax = fig.add_axes([0, 0, 1, 1], frameon=False)    
    ip = InsetPosition(ax, [0, cb_pos, 0.15, cb_height]) #posx, posy, width, height
    cb_bbox_ax.set_axes_locator(ip)
    cb_bbox_ax.xaxis.set_visible(False)
    cb_bbox_ax.yaxis.set_visible(False)  
    
    cb_ax = inset_axes(cb_bbox_ax, width='100%', height='30%', borderpad=0, loc='center left')
    cb = fig.colorbar(hm, cax=cb_ax, orientation = 'horizontal')
    cb_ax.tick_params(labelsize='xx-small', direction='in')
    cb_ax.yaxis.set_label_position("right")
    cb_ax.set_ylabel(cb_label, fontsize='x-small', rotation=0, ha='left', va='center_baseline')
    

    if hasattr(ax.yaxis, 'add_annotation'):
        [ax.yaxis.add_annotation(text, pos) for pos, text in zip(ticks, ticklabels)]
        ax.yaxis.set_ticks([])
    else:
        ax.yaxis.set_ticks(ticks)
        ax.yaxis.set_ticklabels(ticklabels)

    format_axes_to_interval(ax, interval, xaxis=xaxis, yaxis=yaxis)
    clear_spines(ax)

    return ax

# ------------------------

def segment_plot(interval, segments, pad_points=1, ax=None, **kwargs):
    
    if not ax:
        ax = plt.gca()
    fig = ax.figure

    xaxis = kwargs.pop('xaxis', 'bottom')
    yaxis = kwargs.pop('yaxis', None)

    rectprops = {}
    rectprops['color'] = 'k'
    rectprops['edgecolor'] = 'none'
    # rectprops.update(kwargs)

    ax.set_xlim(interval.start, interval.end)

    trans_left = ax.transData + mtransforms.ScaledTranslation(-1 * pad_points / 72., 0,  ax.figure.dpi_scale_trans)
    trans_right = ax.transData + mtransforms.ScaledTranslation(pad_points / 72., 0,  ax.figure.dpi_scale_trans)

    x0 = ax.transData.inverted().transform(trans_left.transform((interval.start, 0)))[0]
    x1 = ax.transData.inverted().transform(trans_right.transform((interval.start, 0)))[0]

    pad_bp = (x1-x0)//2
    row_indices = pack_rows(segments, pad=pad_bp)

    patches = []
    for i, row_index in row_indices.items():
        patches.append(mpatches.Rectangle((i.start, row_index+0.3), i.end-i.start, 0.4, **rectprops))

    pc = mcollections.PatchCollection(patches, match_original=True)
    ax.add_collection(pc)

    ax.set_ylim(min(row_indices.values()), max(row_indices.values())+1)
    ax.invert_yaxis()

    #ax.yaxis.set_ticks(np.array(list(set(row_idxs)))+0.5)
    #ax.yaxis.set_ticklabels([])

    format_axes_to_interval(ax, interval, xaxis=xaxis, yaxis=yaxis)

    return ax

# ------------------------



"""Enables iterable processing of datasets for multithreading
"""
import sys
import multiprocessing
import threading
import traceback

from genome_tools.data.samplers import random_sampler, sequential_sampler, minibatch_sampler
from genome_tools.data.utils import list_collate

import logging
logger = logging.getLogger(__name__)

default_collate = list_collate

class exception_wrapper(object):
    """Wraps an exception plus traceback to communicate across threads"""

    def __init__(self, exc_info):
        self.exc_type = exc_info[0]
        self.exc_msg = "".join(traceback.format_exception(*exc_info))

def _worker_loop(process, index_queue, data_queue, collate_fn):
    """Worker loop to load data"""
    while True:
        r = index_queue.get()
        if r is None:
            break
        idx, batch_indices = r

        try:
            batch = collate_fn([process[i] for i in batch_indices])
        except Exception:
            data_queue.put((idx, exception_wrapper(sys.exc_info())))
        else:
            data_queue.put((idx, batch))

    # clean up function that closes filehandlers, etc.
    if hasattr(process, 'cleanup'):
        process.cleanup()

class data_loader_iter(object):
    """Iterable data loader class
    
    Attributes
    ----------


    """
    def __init__(self, loader):
        """
        

        Parameters
        ----------
        """
        self.process = loader.process
        self.collate_fn = loader.collate_fn
        self.batch_sampler = loader.batch_sampler
        self.num_workers = loader.num_workers
        self.done_event = threading.Event()

        self.sample_iter = iter(self.batch_sampler)
                
        if self.num_workers > 1:

            logger.info(f"Using {self.num_workers} threads to process data")

            self.index_queue = multiprocessing.SimpleQueue()
            self.data_queue = multiprocessing.SimpleQueue()
            self.batches_outstanding = 0
            self.shutdown = False
            self.send_idx = 0
            self.rcvf_idx = 0
            self.reorder_dict = {}

            self.workers = [
                multiprocessing.Process(
                    target=_worker_loop,
                    args=(self.process, self.index_queue, self.data_queue, self.collate_fn))
                for _ in range(self.num_workers)
            ]

            for w in self.workers:
                w.daemon = True
                w.start()

            for _ in range(2 * self.num_workers):
                self._put_indices()
        else:

            logger.info(f"Using a single threads to process data")

    
    def __len__(self):
        return len(self.batch_sampler)

    def __next__(self):
        if self.num_workers <= 1:
            indices = next(self.sample_iter)
            batch = self.collate_fn([self.process[i] for i in indices])
            return batch
        
        if self.rcvf_idx in self.reorder_dict:
            batch = self.reorder_dict.pop(self.rcvf_idx)
            return self._process_next_batch(batch)
        
        if self.batches_outstanding == 0:
            self._shutdown_workers()
            raise StopIteration

        while True:
            assert (not self.shutdown and self.batches_outstanding > 0)
            idx, batch = self.data_queue.get()
            self.batches_outstanding -= 1
            if idx != self.rcvf_idx:
                # store out of order samples
                self.reorder_dict[idx] = batch
                continue
            return self._process_next_batch(batch)
            
    def __iter__(self):
        return self

    def _put_indices(self):
        assert self.batches_outstanding < 2 * self.num_workers
        indices = next(self.sample_iter, None)
        if indices == None:
            return
        self.index_queue.put((self.send_idx, indices))
        self.batches_outstanding += 1
        self.send_idx += 1

    def _process_next_batch(self, batch):
        self.rcvf_idx += 1
        self._put_indices()
        if isinstance(batch, exception_wrapper):
            raise batch.exc_type(batch.exc_msg)
        return batch

    def __getstate__(self):
        raise NotImplementedError("Class cannot be pickled")

    def _shutdown_workers(self):
        if not self.shutdown:
            self.shutdown = True
            self.done_event.set()
            for _ in self.workers:
                self.index_queue.put(None)
                # wait for worker to join?

    def __del__(self):
        if self.num_workers > 0:
            self._shutdown_workers()

class data_loader(object):
    def __init__(self, process, batch_size=1, shuffle=False, num_workers=0, collate_fn=default_collate, sampler=None, batch_sampler=None, drop_last=False):
        self.process = process
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.collate_fn = collate_fn
        
        if batch_sampler is not None:
            if batch_size > 1 or shufffle or sampler is not None or drop_last:
                raise ValueError('batch_sampler is mutually exclusive with batch_size, shuffle, sampler and drop_last')
        
        if sampler is not None and shuffle:
            raise ValueError('sampler is mutually exclusive with shuffle')
        
        if batch_sampler is None:
            if sampler is None:
                if shuffle:
                    sampler = random_sampler(process)
                else:    
                    sampler = sequential_sampler(process)
            batch_sampler = minibatch_sampler(sampler, batch_size, drop_last)

        self.sampler = sampler
        self.batch_sampler = batch_sampler

        logger.info(f"Processing data with batch_size = {self.batch_size:,} resulting in {len(self):,} batches" )
        logger.info(f"Using '{self.collate_fn.__doc__}' to collate batch chunks")

    def __iter__(self):
        return data_loader_iter(self)

    def __len__(self):
        return len(self.batch_sampler)




import numpy as np

import matplotlib.ticker as mticker

__all__ = ["clear_spines", "get_genomic_axis_tick_formatter", "format_axes_to_interval", "rescale_data"]

def clear_spines(ax):
    """Clear spines from plot"""

    for loc, spine in ax.spines.items():
        spine.set_visible(False)

    ax.patch.set_visible(False)


def get_genomic_axis_tick_formatter(interval):
    """Return a tick formattor for the genomic coordinate axis"""

    if len(interval) > 1e5:
        fmt='{:0.2f} Mb'
        denom=1e6
    else:
        fmt='{:,.0f}'
        denom=1

    def _func(x, pos):
        return fmt.format(x/denom) 
    
    return mticker.FuncFormatter(_func)


def format_axes_to_interval(ax, interval, **kwargs):
    """Format axis appearance"""

    ax.set_xlim(interval.start, interval.end)

    ax.tick_params(
        axis='both',
        direction='out'
    )

    ax.xaxis.set(
        major_locator=mticker.MaxNLocator(3, prune='both'),
        minor_locator=mticker.AutoMinorLocator(4),
        major_formatter=get_genomic_axis_tick_formatter(interval),
    )

    ax.yaxis.set(
        major_locator=mticker.MaxNLocator(3, prune='both'),
        minor_locator=mticker.AutoMinorLocator(),
    )


def rescale_data(interval, data, ax, downsample=0, win_fn=np.mean, **kwargs):
    """Stretch and downsample data points for plotting"""

    fig = ax.figure

    w, h = fig.get_size_inches()
    l = len(interval)

    total_pts = fig.get_dpi() * w / (2**downsample) if downsample is not None else l
    stride = int(l//total_pts)

    # Up- or downsample data
    sample_idx = np.linspace(0, len(data)-1, l).astype(int)

    x = np.arange(interval.start, interval.end)
    y = data[sample_idx]

    if downsample > 0:
        x = x[::stride]
        y = [ win_fn(y[i:i+stride]) for i in np.arange(l)[::stride] ]

    assert len(x) == len(y)
    
    return x, y



class row_element:
    def __init__(self, interval, prev=None, next=None, pad=0):
        self.interval = interval
        self.start = interval.start - pad
        self.end = interval.end + pad
        self.prev = prev
        self.next = next

class row:
    def __init__(self, i):
        self.i = i
        self.first = None
        self.last = None

    def add(self, e):

        if self.first is None:
            e.prev = None
            e.next = None
            self.first = self.last = e
            return 1

        curr = self.first
        while curr and curr.start < e.start:
            curr = curr.next

        if curr is None:
            if self.last.end < e.start:
                e.prev = self.last
                e.next = None
                self.last.next = e
                self.last = e
                return 1
            else:
                return 0

        prev = curr.prev
        if prev is None:
            if e.end < curr.start:
                e.prev = None
                e.next = curr
                curr.prev = e
                self.first = e
                return 1
            else:
                return 0

        if prev.end < e.start and curr.start > e.end:
            e.prev = prev
            e.next = curr
            prev.next = e
            curr.prev = e
            return 1
        
        return 0

def pack_rows(intervals, pad=5):
    rows = []
    row_indices = {}
    curr_row = -1

    for interval in intervals:
        
        e = row_element(interval, pad=pad)

        placed = False 

        for r in rows:
            if r.add(e):
                row_indices[e.interval] = r.i
                placed = True
                break
        
        if placed:
            continue

        curr_row += 1
        r = row(curr_row)
        r.add(e)
        rows.append(r)
        row_indices[e.interval] = r.i

    assert len(row_indices) == len(intervals)
    return row_indices


import numpy as np

import matplotlib.pyplot as plt
import matplotlib.collections as mcollections
import matplotlib.patches as mpatches

from .utils import clear_spines

from genome_tools.plotting.colors.cm import get_vocab_color
from genome_tools.helpers import open_file

def read_ideogram(filepath):
    
    xranges={}
    colors={}
    centromeres={}

    with open_file(filepath) as f:

        last_chrom=None
        xr=[]
        for line in f:
            chrom, start, stop, label, stain = line.strip().split('\t')
            start = int(start)
            stop = int(stop)
            width = stop - start
        
            if stain=="acen":
                centromeres[chrom] = centromeres.get(chrom, []) + [(start, width)]
                continue
        
            xranges[chrom] = xranges.get(chrom, []) + [(start, width)]
            colors[chrom] = colors.get(chrom, []) + [(get_vocab_color(stain, 'ideogram'))]

    xranges = xranges
    colors = colors
    centromeres = centromeres

    return xranges, colors, centromeres

def ideogram_plot(data, chrom, pos=None, ax=None, **kwargs):

    if not ax:
        ax = plt.gca()
    fig = ax.figure

    try:
        xranges = data[0][chrom]
        colors = data[1][chrom]
        centromeres = data[2][chrom]
    except:
        print("Error: No chromosome named: {}".format(chrom))
        return

    yranges = (0, 0.5)

    coll = mcollections.BrokenBarHCollection(xranges, yranges, facecolors=colors, edgecolors='black', linewidths=0.5)    
    ax.add_collection(coll)

    if pos:
        ax.axvline(pos, color='red', lw=4)
    w = xranges[-1][0]+xranges[-1][1]

    pad=w*0.05

    ax.set_xlim(0-pad, xranges[-1][0]+xranges[-1][1]+pad)
    ax.xaxis.set_visible(False)

    center = yranges[0] + yranges[1]/2.

    x0, y0 = centromeres[0][0], yranges[0]
    x1, y1 = centromeres[0][0], yranges[1]
    x2, y2 = centromeres[0][0]+centromeres[0][1], center
    cent = mpatches.Polygon(np.array([[x0, y0], [x1, y1], [x2, y2]]), closed=True, 
                    fc=get_vocab_color('acen', 'ideogram'), ec='black', linewidth=0.5)
    ax.add_patch(cent)

    x0, y0 = centromeres[1][0], center
    x1, y1 = centromeres[1][0] + centromeres[1][1], yranges[1]
    x2, y2 = centromeres[1][0] + centromeres[1][1], yranges[0]

    cent = mpatches.Polygon(np.array([[x0, y0], [x1, y1], [x2, y2]]), closed=True,
                    fc=get_vocab_color('acen', 'ideogram'), ec='black', linewidth=0.5)
    ax.add_patch(cent)

    ax.set_yticks([center])
    ax.set_yticklabels([chrom])
    ax.set_ylim(-0.2, 0.7)

    clear_spines(ax)



import numpy as np
import collections
import itertools

# ------------------------

string_classes = (str, bytes)

def _list_collate(stack_fn=list):
    def list_collate_fn(batch):
        if isinstance(batch[0], collections.Mapping):
            return {key: list_collate_fn([d[key] for d in batch]) for key in batch[0]}
        elif isinstance(batch[0], collections.Sequence):
            transposed = zip(*batch)
            return [list_collate_fn(samples) for samples in transposed]
        else:
            return stack_fn(batch)

    return list_collate_fn

list_collate = _list_collate(list)
list_collate.__doc__ = __name__ + '.list_collate'

list_collate_concat = _list_collate(lambda x: list(itertools.chain(*x)))
list_collate_concat.__doc__ = __name__ + '.list_collate_concat'

def _numpy_collate(stack_fn=np.stack):
    def numpy_collate_fn(batch):
        "Puts each data field into a tensor with outer dimension batch size"
        if type(batch[0]).__module__ == 'numpy':
            elem = batch[0]
            if type(elem).__name__ == 'ndarray':
                return stack_fn(batch, 0)
            if elem.shape == ():  # scalars
                return np.array(batch)
        elif isinstance(batch[0], int):
            return np.asarray(batch)
        elif isinstance(batch[0], float):
            return np.asarray(batch)
        elif batch[0] is None:
            return np.asarray(batch)
        elif isinstance(batch[0], string_classes):
            # Also convert to a numpy array
            return np.asarray(batch)
            # return batch
        elif isinstance(batch[0], collections.Mapping):
            return {key: numpy_collate_fn([d[key] for d in batch]) for key in batch[0]}
        elif isinstance(batch[0], collections.Sequence):
            transposed = zip(*batch)
            return [numpy_collate_fn(samples) for samples in transposed]

        raise TypeError(("batch must contain tensors, numbers, dicts or lists; found {}"
                         .format(type(batch[0]))))

    return numpy_collate_fn

numpy_collate = _numpy_collate(np.stack)
numpy_collate.__doc__ = __name__ + '.numpy_collate'

numpy_collate_concat = _numpy_collate(np.concatenate)
numpy_collate_concat.__doc__ = __name__ + '.numpy_collate_concat'

# ------------------------

def check_input_dimensions(intervals):
    """Checks that each interval is of the same width"""
    sizes = list(map(len, intervals))
    if len(np.unique(sizes)) != 1:
        raise IndexError('Inputs are not of same length!')



from __future__ import absolute_import, division, print_function

import os
import sys

from setuptools import find_packages, setup

__version__ = "1.0.4"

install_requires = ["numpy",
					"scipy",
					"pysam",
					"openpyxl",
					"pyBigWig",
					"matplotlib",
					"descartes",
					"shapley"] 

setup(
	name = "genome_tools",
	version = __version__,
	license = "GPL-3.0-or-later",
	description = "A toolkit for processing and plotting genomic datasets.",
	long_description = "",
	author = "Jeff Vierstra",
	author_email = "jvierstra@altius.org",
	zip_safe = False,
	packages =  find_packages(),
    install_requires = install_requires,
    download_url = "https://github.com/jvierstra/genome-tools/archive/v1.0.4.tar.gz",
    classifiers=[
	    'Development Status :: 5 - Production/Stable', 
	    'Intended Audience :: Science/Research', 
	    'Topic :: Scientific/Engineering :: Bio-Informatics',
	    'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
	    'Programming Language :: Python :: 3',
		'Programming Language :: Python :: 3.6',
		'Programming Language :: Python :: 3.7',
		'Programming Language :: Python :: 3.8',
		'Programming Language :: Python :: 3.9',
],
)



"""General class structure to load and process datasets
"""

from genome_tools.data.loaders import data_loader

import logging
logger = logging.getLogger(__name__)

def iterable_cycle(iterable):
    """
    Args:
      iterable: object with an __iter__ method that can be called multiple times
    """
    while True:
        for x in iterable:
            yield x

class base_dataset(object):
    def batch_iter(self, **kwargs):
        raise  NotImplementedError

    def batch_train_iter(self, cycle=True, **kwargs):
        
        if cycle:
            return ((x["inputs"], x["targets"])
                    for x in iterable_cycle(self._batch_iterable(**kwargs)))
        else:
            return ((x["inputs"], x["targets"]) 
                    for x in self.batch_iter(**kwargs))
        
    def batch_predict_iter(self, cycle=True, **kwargs):
        
        return ((x["inputs"], x["targets"]) 
                    for x in self.batch_iter(**kwargs))
        
    def load_all(self, **kwargs):
        """Load an entire dataset"""
        return [x for x in self.batch_iter(**kwargs)]
        
class dataset(base_dataset):
    """All datasets should subclass this class. All subclases should
    override `__len__` and `__getitem__` to support integer indexing"""

    def __getitem__(self, index):
        """Return one sample"""
        raise NotImplementedError
    
    def __len__(self):
        """Return number of all elements"""
        raise NotImplementedError

    def _batch_iterable(self, batch_size=1, num_workers=0, **kwargs):
        
        dp = data_loader(self, 
                        batch_size=batch_size,
                        num_workers=num_workers,
                        **kwargs)

        return dp

    def batch_iter(self, batch_size=1, num_workers=0, **kwargs):
        
        dp =  self._batch_iterable(batch_size=batch_size,
                                    num_workers=num_workers,
                                    **kwargs)
        
        return iter(dp)

    # Example of overloaded `load_all` function
    # def load_all(self, batch_size=1, **kwargs):
    #     """Load all data"""
    #     return numpy_collate_concat([x for x in tqdm(self.batch_iter(batch_size, **kwargs))])
    


"""Classes to extract genomics data from different file formats

Note that many of these filehandlers are not thread safe,
so they must be opened separately on different threads when
using multiprocessing.
"""
import numpy as np
import pysam
import pyBigWig as pbw
import pandas as pd

class base_extractor(object):
    def __init__(self, filename, **kwargs):
        self.filename = filename

    def __getitem__(self, i):
        raise NotImplementedError 

# ------------------------

class fasta_extractor(base_extractor):
    def __init__(self, filename, **kwargs):
        super(fasta_extractor, self).__init__(filename, **kwargs)
        
        self.fasta = pysam.FastaFile(filename)

    def __getitem__(self, interval):
        seq = self.fasta.fetch(interval.chrom, interval.start, interval.end)
        # implement any post processing steps here
        return seq

    def __del__(self):
        self.close()

    def close(self):
        if self.fasta and self.fasta.is_open():
            self.fasta.close()

# ------------------------

class tabix_iter(object):
    """Wraps tabix fetch to return an iterator that can be used with pandas"""
    def __init__(self, tabix, interval):
        self.iter = tabix.fetch(interval.chrom, interval.start, interval.end)
    def read(self, n=0):
        try:
            return next(self.iter) + "\n"
        except StopIteration:
            return ''
    def __iter__(self):
        return self
    def __next__(self):
        return self.read()

class tabix_extractor(base_extractor):
    def __init__(self, filename, **kwargs):
        """
        """
        super(tabix_extractor, self).__init__(filename, **kwargs)

        self.tabix = pysam.TabixFile(filename)

    def __getitem__(self, interval):
        ret = pd.read_table(tabix_iter(self.tabix, interval), header=None, index_col=None)
        return ret

    def __del__(self):
        self.close()

    def close(self):
        if self.tabix and self.tabix.is_open():
            self.tabix.close()
    
# ------------------------

class bigwig_extractor(base_extractor):
    def __init__(self, filename, **kwargs):
        """
        """
        super(bigwig_extractor, self).__init__(filename, **kwargs)

        self.bw = pbw.open(filename)

    def __getitem__(self, interval):
        out = self.bw.values(interval.chrom, interval.start, interval.end, numpy=True)
        return out

    def __del__(self):
        self.close()

    def close(self):
        if self.bw:
            self.bw.close()

# ------------------------

class d4_extractor(base_extractor):
    def __init__(self, filename, **kwargs):
        super().__init__(filename, **kwargs)

    def __getitem__(self, i):
        raise NotImplementedError

    def __del__(self):
        self.close()

    def close(self):
        pass

    